#Create a simulator object
set ns [new Simulator]


#sets the start time of the experiment
set START_TIME 0.0

#sets the stop time of the experiment
set STOP_TIME 50.0


#Open a trace file
set nf [open out.nam w]
$ns namtrace-all $nf


#Define a 'finish' procedure
proc finish {} {
        global ns nf
        $ns flush-trace
        close $nf
        #exec nam out.nam &
        exit 0
}



#structure of the network
#
#  [0]-[1]
#


#Create ten nodes

set n0  [$ns node]
set n1  [$ns node]

#Connect the nodes with links
$ns simplex-link $n0 $n1 100Mb 0.1ms dsRED/edge
$ns simplex-link $n1 $n0 100Mb 0.1ms dsRED/edge

set n0n1Link [$ns link $n0 $n1]
set n1n0Link [$ns link $n1 $n0]

#queue qreation
set q01 [$n0n1Link queue]
set q10 [$n1n0Link queue]

##########################################################

$q01 setSchedularMode MDRR_STR
$q01 meanPktSize 462

$q01 set numQueues_ 2
$q01 setNumPrec 2

#background traffic DSCP 0
$q01 addPolicyEntry 0 Dumb 0
$q01 addPolicyEntry 255 Dumb 0
$q01 addPolicerEntry Dumb 0 0

#control packets DSCP 1
$q01 addPolicyEntry 1 TokenBucket 1 25000000 5000
$q01 addPolicerEntry TokenBucket 1 1

#DSCP 0 natural queue 1 virtual queue 0
#DSCP 1 natural queue 0 virtual queue 0
#all others DSCP go to quue 0 virtual queue 1
$q01 addPHBEntry 0 1 0
$q01 addPHBEntry 1 0 0

#$q01 addPHBEntry 2 0 1
#$q01 addPHBEntry 3 0 1
#$q01 addPHBEntry 4 0 1


$q01 addQueueWeights 0 0
$q01 addQueueWeights 1 100

##########################################################

$q10 setSchedularMode MDRR_STR
$q10 meanPktSize 462

$q10 set numQueues_ 2
$q10 setNumPrec 2

#background traffic DSCP 0
$q10 addPolicyEntry 0 Dumb 0
$q10 addPolicyEntry 255 Dumb 0
$q10 addPolicerEntry Dumb 0 0

#control packets DSCP 1
$q10 addPolicyEntry 1 TokenBucket 1 25000000 5000
$q10 addPolicerEntry TokenBucket 1 1

#DSCP 0 natural queue 1 virtual queue 0
#DSCP 1 natural queue 0 virtual queue 0
#all others DSCP go to quue 0 virtual queue 1
$q10 addPHBEntry 0 1 0
$q10 addPHBEntry 1 0 0


$q10 addQueueWeights 0 0
$q10 addQueueWeights 1 100

############################################################


#Define a 'recv' function for the class 'Agent/BBbase'

Agent/BBedge instproc recv {endnode answer dscp bndwdth start_time stop_time} {

 global ns n0 n1 q01 q10

 $self instvar node_
 set sink_node [$ns get-node-by-id $endnode]

 #if positive answer to the bandwidth request
 if {$answer == 1} {
    set sink_node [$ns get-node-by-id $endnode]
    puts "[$node_ id]->[$sink_node id]:The answer was POSITIVE, dscp=$dscp"
    puts "Accepted bandwidth :$bndwdth"

    #here we set the bandwidth of the TokenBucket
    set b2 [expr 1.0*$bndwdth]
    #puts "b2=$b2"

    #set the queue policy on the first node, the marked packet reaches
    #$ns compute-routes
    #set rlg [$ns get-routelogic]
    #set nbghr [$rlg lookup [$node_ id] [$sink_node id]]
    $q01 addPolicyEntry $dscp TokenBucket $dscp $b2 5000
    $q10 addPolicyEntry $dscp TokenBucket $dscp $b2 5000

    $q01 addPolicerEntry TokenBucket $dscp $dscp
    $q10 addPolicerEntry TokenBucket $dscp $dscp

    #all the new DSCP go to queue 0 virtual queue 1
    $q01 addPHBEntry $dscp 0 1
    $q10 addPHBEntry $dscp 0 1

    #create cbr traffic from $node_ to $sink_node and mark the packets with the $dscp number
    #set bndwdth [expr 2*$bndwdth] #we use this if we want to send more bndwdth than we requested
    cbr_traffic $node_ $sink_node $dscp $bndwdth

    } else {

        puts "[$node_ id]->[$sink_node id]:The answer was NEGATIVE, dscp=$dscp"
        puts "Rejected bandwidth :$bndwdth"
    }

}




#every node gets a BBedge agent

set p0 [new Agent/BBedge]
$ns attach-agent $n0 $p0

set p1 [new Agent/BBedge]
$ns attach-agent $n1 $p1


#BBbase agent goes to node 0
set p44 [new Agent/BBbase/CentralAacr]
$ns attach-agent $n0 $p44


#Connect all the BBedge agents with the BBbase agent
$ns connect $p44 $p0
$ns connect $p44 $p1

#every BBedge agent controls 10Mbps of every link
$p0 set_bndw 1 10000000 $START_TIME $STOP_TIME
$p1 set_bndw 0 10000000 $START_TIME $STOP_TIME


#######CBR TRAFFIC ########################
#function cbr_traffic creates udp traffic with constant rate
proc cbr_traffic {source_node dest_node dscp bndwdth} {

 global ns START_TIME STOP_TIME
	set cbr_agent [new Agent/TCP]
	$ns attach-agent $source_node $cbr_agent
	$cbr_agent classify $dscp

	set cbr_sink [new Agent/TCPSink]
	$cbr_sink classify $dscp
	$ns attach-agent $dest_node $cbr_sink

	set cbr_source [new Application/Traffic/CBR]
	$cbr_source attach-agent $cbr_agent
	$cbr_source set rate_ 200kbps

	$ns connect $cbr_agent $cbr_sink


        $cbr_source start

	$ns at $STOP_TIME "$cbr_source stop"

}

$ns at 0.1 "$p0 inform $START_TIME $STOP_TIME"
$ns at 0.1 "$p1 inform $START_TIME $STOP_TIME"

#Insert extra files
source rnd_rqst_PBAC_1.tcl

#Notify BBbase that the experiment is ending, to examine requests left in queue
$ns at $STOP_TIME "$p44 finish"

$ns at $STOP_TIME "finish"


#Run the simulation

$ns run


